@@ -18,13 +18,14 @@ package core
 
 import (
 	"fmt"
+	"k8s.io/apimachinery/pkg/types"
 	"math/rand"
 	"reflect"
-	"sort"
 	"time"
 
-	"k8s.io/kubernetes/pkg/scheduler/util"
-
+	appsv1 "k8s.io/api/apps/v1"
+	apiv1 "k8s.io/api/core/v1"
+	apiequality "k8s.io/apimachinery/pkg/api/equality"
 	"k8s.io/autoscaler/cluster-autoscaler/cloudprovider"
 	"k8s.io/autoscaler/cluster-autoscaler/clusterstate"
 	"k8s.io/autoscaler/cluster-autoscaler/clusterstate/utils"
@@ -38,11 +39,7 @@ import (
 	"k8s.io/autoscaler/cluster-autoscaler/utils/glogx"
 	"k8s.io/autoscaler/cluster-autoscaler/utils/gpu"
 	kube_util "k8s.io/autoscaler/cluster-autoscaler/utils/kubernetes"
-	scheduler_util "k8s.io/autoscaler/cluster-autoscaler/utils/scheduler"
-
-	appsv1 "k8s.io/api/apps/v1"
-	apiv1 "k8s.io/api/core/v1"
-	apiequality "k8s.io/apimachinery/pkg/api/equality"
+	schedulerapi "k8s.io/kubernetes/pkg/scheduler/api"
 	schedulernodeinfo "k8s.io/kubernetes/pkg/scheduler/nodeinfo"
 
 	"k8s.io/klog"
@@ -53,6 +50,20 @@ const (
 	ReschedulerTaintKey = "CriticalAddonsOnly"
 )
 
+var (
+	nodeConditionTaints = taintKeySet{
+		schedulerapi.TaintNodeNotReady:           true,
+		schedulerapi.TaintNodeUnreachable:        true,
+		schedulerapi.TaintNodeUnschedulable:      true,
+		schedulerapi.TaintNodeMemoryPressure:     true,
+		schedulerapi.TaintNodeDiskPressure:       true,
+		schedulerapi.TaintNodeNetworkUnavailable: true,
+		schedulerapi.TaintNodePIDPressure:        true,
+		schedulerapi.TaintExternalCloudProvider:  true,
+		schedulerapi.TaintNodeShutdown:           true,
+	}
+)
+
 // Following data structure is used to avoid running predicates #pending_pods * #nodes
 // times (which turned out to be very expensive if there are thousands of pending pods).
 // This optimization is based on the assumption that if there are that many pods they're
@@ -75,6 +86,8 @@ type podSchedulableInfo struct {
 
 type podSchedulableMap map[string][]podSchedulableInfo
 
+type taintKeySet map[string]bool
+
 func (psi *podSchedulableInfo) match(pod *apiv1.Pod) bool {
 	return reflect.DeepEqual(pod.Labels, psi.labels) && apiequality.Semantic.DeepEqual(pod.Spec, psi.spec)
 }
@@ -108,75 +121,6 @@ func (podMap podSchedulableMap) set(pod *apiv1.Pod, err *simulator.PredicateErro
 	})
 }
 
-// filterOutSchedulableByPacking checks whether pods from <unschedulableCandidates> marked as unschedulable
-// can be scheduled on free capacity on existing nodes by trying to pack the pods. It tries to pack the higher priority
-// pods first. It takes into account pods that are bound to node and will be scheduled after lower priority pod preemption.
-func filterOutSchedulableByPacking(unschedulableCandidates []*apiv1.Pod, nodes []*apiv1.Node, allScheduled []*apiv1.Pod, podsWaitingForLowerPriorityPreemption []*apiv1.Pod,
-	predicateChecker *simulator.PredicateChecker, expendablePodsPriorityCutoff int) []*apiv1.Pod {
-	var unschedulablePods []*apiv1.Pod
-	nonExpendableScheduled := filterOutExpendablePods(allScheduled, expendablePodsPriorityCutoff)
-	nodeNameToNodeInfo := scheduler_util.CreateNodeNameToInfoMap(append(nonExpendableScheduled, podsWaitingForLowerPriorityPreemption...), nodes)
-	loggingQuota := glogx.PodsLoggingQuota()
-
-	sort.Slice(unschedulableCandidates, func(i, j int) bool {
-		return util.GetPodPriority(unschedulableCandidates[i]) > util.GetPodPriority(unschedulableCandidates[j])
-	})
-
-	for _, pod := range unschedulableCandidates {
-		nodeName, err := predicateChecker.FitsAny(pod, nodeNameToNodeInfo)
-		if err != nil {
-			unschedulablePods = append(unschedulablePods, pod)
-		} else {
-			glogx.V(4).UpTo(loggingQuota).Infof("Pod %s marked as unschedulable can be scheduled on %s. Ignoring in scale up.", pod.Name, nodeName)
-			nodeNameToNodeInfo[nodeName] = scheduler_util.NodeWithPod(nodeNameToNodeInfo[nodeName], pod)
-		}
-	}
-
-	glogx.V(4).Over(loggingQuota).Infof("%v other pods marked as unschedulable can be scheduled.", -loggingQuota.Left())
-	return unschedulablePods
-}
-
-// filterOutSchedulableSimple checks whether pods from <unschedulableCandidates> marked as unschedulable
-// by Scheduler actually can't be scheduled on any node and filter out the ones that can.
-// It takes into account pods that are bound to node and will be scheduled after lower priority pod preemption.
-func filterOutSchedulableSimple(unschedulableCandidates []*apiv1.Pod, nodes []*apiv1.Node, allScheduled []*apiv1.Pod, podsWaitingForLowerPriorityPreemption []*apiv1.Pod,
-	predicateChecker *simulator.PredicateChecker, expendablePodsPriorityCutoff int) []*apiv1.Pod {
-	var unschedulablePods []*apiv1.Pod
-	nonExpendableScheduled := filterOutExpendablePods(allScheduled, expendablePodsPriorityCutoff)
-	nodeNameToNodeInfo := scheduler_util.CreateNodeNameToInfoMap(append(nonExpendableScheduled, podsWaitingForLowerPriorityPreemption...), nodes)
-	podSchedulable := make(podSchedulableMap)
-	loggingQuota := glogx.PodsLoggingQuota()
-
-	for _, pod := range unschedulableCandidates {
-		cachedError, found := podSchedulable.get(pod)
-		// Try to get result from cache.
-		if found {
-			if cachedError != nil {
-				unschedulablePods = append(unschedulablePods, pod)
-			} else {
-				glogx.V(4).UpTo(loggingQuota).Infof("Pod %s marked as unschedulable can be scheduled (based on simulation run for other pod owned by the same controller). Ignoring in scale up.", pod.Name)
-			}
-			continue
-		}
-
-		// Not found in cache, have to run the predicates.
-		nodeName, err := predicateChecker.FitsAny(pod, nodeNameToNodeInfo)
-		// err returned from FitsAny isn't a PredicateError.
-		// Hello, ugly hack. I wish you weren't here.
-		var predicateError *simulator.PredicateError
-		if err != nil {
-			predicateError = simulator.NewPredicateError("FitsAny", err, nil, nil)
-			unschedulablePods = append(unschedulablePods, pod)
-		} else {
-			glogx.V(4).UpTo(loggingQuota).Infof("Pod %s marked as unschedulable can be scheduled on %s. Ignoring in scale up.", pod.Name, nodeName)
-		}
-		podSchedulable.set(pod, predicateError)
-	}
-
-	glogx.V(4).Over(loggingQuota).Infof("%v other pods marked as unschedulable can be scheduled.", -loggingQuota.Left())
-	return unschedulablePods
-}
-
 // filterOutExpendableAndSplit filters out expendable pods and splits into:
 //   - waiting for lower priority pods preemption
 //   - other pods.
@@ -207,40 +151,89 @@ func filterOutExpendablePods(pods []*apiv1.Pod, expendablePodsPriorityCutoff int
 	return result
 }
 
-// checkPodsSchedulableOnNode checks if pods can be scheduled on the given node.
-func checkPodsSchedulableOnNode(context *context.AutoscalingContext, pods []*apiv1.Pod, nodeGroupId string, nodeInfo *schedulernodeinfo.NodeInfo) map[*apiv1.Pod]*simulator.PredicateError {
-	schedulingErrors := map[*apiv1.Pod]*simulator.PredicateError{}
-	loggingQuota := glogx.PodsLoggingQuota()
-	podSchedulable := make(podSchedulableMap)
+type equivalenceGroupId int
+
+type podsSchedulableOnNodeChecker struct {
+	context               *context.AutoscalingContext
+	pods                  []*apiv1.Pod
+	podsEquivalenceGroups map[types.UID]equivalenceGroupId
+}
+
+func newPodsSchedulableOnNodeChecker(context *context.AutoscalingContext, pods []*apiv1.Pod) *podsSchedulableOnNodeChecker {
+	checker := podsSchedulableOnNodeChecker{
+		context:               context,
+		pods:                  pods,
+		podsEquivalenceGroups: make(map[types.UID]equivalenceGroupId),
+	}
+
+	// compute the podsEquivalenceGroups
+	var nextGroupId equivalenceGroupId
+	type equivalanceGroup struct {
+		id           equivalenceGroupId
+		representant *apiv1.Pod
+	}
+
+	equivalenceGroupsByController := make(map[types.UID][]equivalanceGroup)
 
 	for _, pod := range pods {
-		// Check if pod isn't repeated before overwriting result for it.
-		if _, repeated := schedulingErrors[pod]; repeated {
-			// This shouldn't really happen.
-			klog.Warningf("Pod %v appears multiple time on pods list, will only count it once in scale-up simulation", pod)
+		controllerRef := drain.ControllerRef(pod)
+		if controllerRef == nil {
+			checker.podsEquivalenceGroups[pod.UID] = nextGroupId
+			nextGroupId++
+			continue
 		}
-		// Try to get result from cache.
-		err, found := podSchedulable.get(pod)
-		if found {
-			schedulingErrors[pod] = err
-			if err != nil {
-				glogx.V(2).UpTo(loggingQuota).Infof("Pod %s can't be scheduled on %s. Used cached predicate check results", pod.Name, nodeGroupId)
+
+		matchingFound := false
+		for _, g := range equivalenceGroupsByController[controllerRef.UID] {
+			if reflect.DeepEqual(pod.Labels, g.representant.Labels) && apiequality.Semantic.DeepEqual(pod.Spec, g.representant.Spec) {
+				matchingFound = true
+				checker.podsEquivalenceGroups[pod.UID] = g.id
+				break
 			}
 		}
+
+		if !matchingFound {
+			newGroup := equivalanceGroup{
+				id:           nextGroupId,
+				representant: pod,
+			}
+			equivalenceGroupsByController[controllerRef.UID] = append(equivalenceGroupsByController[controllerRef.UID], newGroup)
+			checker.podsEquivalenceGroups[pod.UID] = newGroup.id
+			nextGroupId++
+		}
+	}
+
+	return &checker
+}
+
+// checkPodsSchedulableOnNode checks if pods can be scheduled on the given node.
+func (c *podsSchedulableOnNodeChecker) checkPodsSchedulableOnNode(nodeGroupId string, nodeInfo *schedulernodeinfo.NodeInfo) map[*apiv1.Pod]*simulator.PredicateError {
+	loggingQuota := glogx.PodsLoggingQuota()
+	schedulingErrors := make(map[equivalenceGroupId]*simulator.PredicateError)
+
+	for _, pod := range c.pods {
+		equivalenceGroup := c.podsEquivalenceGroups[pod.UID]
+		err, found := schedulingErrors[equivalenceGroup]
+		if found && err != nil {
+			glogx.V(2).UpTo(loggingQuota).Infof("Pod %s can't be scheduled on %s. Used cached predicate check results", pod.Name, nodeGroupId)
+		}
 		// Not found in cache, have to run the predicates.
 		if !found {
-			err = context.PredicateChecker.CheckPredicates(pod, nil, nodeInfo)
-			podSchedulable.set(pod, err)
-			schedulingErrors[pod] = err
+			err = c.context.PredicateChecker.CheckPredicates(pod, nil, nodeInfo)
+			schedulingErrors[equivalenceGroup] = err
 			if err != nil {
 				// Always log for the first pod in a controller.
 				klog.V(2).Infof("Pod %s can't be scheduled on %s, predicate failed: %v", pod.Name, nodeGroupId, err.VerboseError())
 			}
 		}
 	}
-
 	glogx.V(2).Over(loggingQuota).Infof("%v other pods can't be scheduled on %s.", -loggingQuota.Left(), nodeGroupId)
-	return schedulingErrors
+
+	schedulingErrorsByPod := make(map[*apiv1.Pod]*simulator.PredicateError)
+	for _, pod := range c.pods {
+		schedulingErrorsByPod[pod] = schedulingErrors[c.podsEquivalenceGroups[pod.UID]]
+	}
+	return schedulingErrorsByPod
 }
 
 // getNodeInfosForGroups finds NodeInfos for all node groups used to manage the given nodes. It also returns a node group to sample node mapping.
@@ -248,10 +241,15 @@ func checkPodsSchedulableOnNode(context *context.AutoscalingContext, pods []*api
 //
 // TODO(mwielgus): Review error policy - sometimes we may continue with partial errors.
 func getNodeInfosForGroups(nodes []*apiv1.Node, nodeInfoCache map[string]*schedulernodeinfo.NodeInfo, cloudProvider cloudprovider.CloudProvider, listers kube_util.ListerRegistry,
-	daemonsets []*appsv1.DaemonSet, predicateChecker *simulator.PredicateChecker) (map[string]*schedulernodeinfo.NodeInfo, errors.AutoscalerError) {
+	daemonsets []*appsv1.DaemonSet, predicateChecker *simulator.PredicateChecker, ignoredTaints taintKeySet) (map[string]*schedulernodeinfo.NodeInfo, errors.AutoscalerError) {
 	result := make(map[string]*schedulernodeinfo.NodeInfo)
 	seenGroups := make(map[string]bool)
 
+	podsForNodes, err := getPodsForNodes(listers)
+	if err != nil {
+		return map[string]*schedulernodeinfo.NodeInfo{}, err
+	}
+
 	// processNode returns information whether the nodeTemplate was generated and if there was an error.
 	processNode := func(node *apiv1.Node) (bool, string, errors.AutoscalerError) {
 		nodeGroup, err := cloudProvider.NodeGroupForNode(node)
@@ -264,11 +262,11 @@ func getNodeInfosForGroups(nodes []*apiv1.Node, nodeInfoCache map[string]*schedu
 		id := nodeGroup.Id()
 		if _, found := result[id]; !found {
 			// Build nodeInfo.
-			nodeInfo, err := simulator.BuildNodeInfoForNode(node, listers)
+			nodeInfo, err := simulator.BuildNodeInfoForNode(node, podsForNodes)
 			if err != nil {
 				return false, "", err
 			}
-			sanitizedNodeInfo, err := sanitizeNodeInfo(nodeInfo, id)
+			sanitizedNodeInfo, err := sanitizeNodeInfo(nodeInfo, id, ignoredTaints)
 			if err != nil {
 				return false, "", err
 			}
@@ -312,7 +310,7 @@ func getNodeInfosForGroups(nodes []*apiv1.Node, nodeInfoCache map[string]*schedu
 
 		// No good template, trying to generate one. This is called only if there are no
 		// working nodes in the node groups. By default CA tries to use a real-world example.
-		nodeInfo, err := getNodeInfoFromTemplate(nodeGroup, daemonsets, predicateChecker)
+		nodeInfo, err := getNodeInfoFromTemplate(nodeGroup, daemonsets, predicateChecker, ignoredTaints)
 		if err != nil {
 			if err == cloudprovider.ErrNotImplemented {
 				continue
@@ -353,8 +351,20 @@ func getNodeInfosForGroups(nodes []*apiv1.Node, nodeInfoCache map[string]*schedu
 	return result, nil
 }
 
+func getPodsForNodes(listers kube_util.ListerRegistry) (map[string][]*apiv1.Pod, errors.AutoscalerError) {
+	pods, err := listers.ScheduledPodLister().List()
+	if err != nil {
+		return nil, errors.ToAutoscalerError(errors.ApiCallError, err)
+	}
+	podsForNodes := map[string][]*apiv1.Pod{}
+	for _, p := range pods {
+		podsForNodes[p.Spec.NodeName] = append(podsForNodes[p.Spec.NodeName], p)
+	}
+	return podsForNodes, nil
+}
+
 // getNodeInfoFromTemplate returns NodeInfo object built base on TemplateNodeInfo returned by NodeGroup.TemplateNodeInfo().
-func getNodeInfoFromTemplate(nodeGroup cloudprovider.NodeGroup, daemonsets []*appsv1.DaemonSet, predicateChecker *simulator.PredicateChecker) (*schedulernodeinfo.NodeInfo, errors.AutoscalerError) {
+func getNodeInfoFromTemplate(nodeGroup cloudprovider.NodeGroup, daemonsets []*appsv1.DaemonSet, predicateChecker *simulator.PredicateChecker, ignoredTaints taintKeySet) (*schedulernodeinfo.NodeInfo, errors.AutoscalerError) {
 	id := nodeGroup.Id()
 	baseNodeInfo, err := nodeGroup.TemplateNodeInfo()
 	if err != nil {
@@ -365,7 +375,7 @@ func getNodeInfoFromTemplate(nodeGroup cloudprovider.NodeGroup, daemonsets []*ap
 	pods = append(pods, baseNodeInfo.Pods()...)
 	fullNodeInfo := schedulernodeinfo.NewNodeInfo(pods...)
 	fullNodeInfo.SetNode(baseNodeInfo.Node())
-	sanitizedNodeInfo, typedErr := sanitizeNodeInfo(fullNodeInfo, id)
+	sanitizedNodeInfo, typedErr := sanitizeNodeInfo(fullNodeInfo, id, ignoredTaints)
 	if typedErr != nil {
 		return nil, typedErr
 	}
@@ -403,9 +413,9 @@ func deepCopyNodeInfo(nodeInfo *schedulernodeinfo.NodeInfo) (*schedulernodeinfo.
 	return newNodeInfo, nil
 }
 
-func sanitizeNodeInfo(nodeInfo *schedulernodeinfo.NodeInfo, nodeGroupName string) (*schedulernodeinfo.NodeInfo, errors.AutoscalerError) {
+func sanitizeNodeInfo(nodeInfo *schedulernodeinfo.NodeInfo, nodeGroupName string, ignoredTaints taintKeySet) (*schedulernodeinfo.NodeInfo, errors.AutoscalerError) {
 	// Sanitize node name.
-	sanitizedNode, err := sanitizeTemplateNode(nodeInfo.Node(), nodeGroupName)
+	sanitizedNode, err := sanitizeTemplateNode(nodeInfo.Node(), nodeGroupName, ignoredTaints)
 	if err != nil {
 		return nil, err
 	}
@@ -426,7 +436,7 @@ func sanitizeNodeInfo(nodeInfo *schedulernodeinfo.NodeInfo, nodeGroupName string
 	return sanitizedNodeInfo, nil
 }
 
-func sanitizeTemplateNode(node *apiv1.Node, nodeGroup string) (*apiv1.Node, errors.AutoscalerError) {
+func sanitizeTemplateNode(node *apiv1.Node, nodeGroup string, ignoredTaints taintKeySet) (*apiv1.Node, errors.AutoscalerError) {
 	newNode := node.DeepCopy()
 	nodeName := fmt.Sprintf("template-node-for-%s-%d", nodeGroup, rand.Int63())
 	newNode.Labels = make(map[string]string, len(node.Labels))
@@ -446,13 +456,27 @@ func sanitizeTemplateNode(node *apiv1.Node, nodeGroup string) (*apiv1.Node, erro
 		switch taint.Key {
 		case ReschedulerTaintKey:
 			klog.V(4).Infof("Removing rescheduler taint when creating template from node %s", node.Name)
+			continue
 		case deletetaint.ToBeDeletedTaint:
 			klog.V(4).Infof("Removing autoscaler taint when creating template from node %s", node.Name)
+			continue
 		case deletetaint.DeletionCandidateTaint:
 			klog.V(4).Infof("Removing autoscaler soft taint when creating template from node %s", node.Name)
-		default:
-			newTaints = append(newTaints, taint)
+			continue
 		}
+
+		// ignore conditional taints as they represent a transient node state.
+		if exists := nodeConditionTaints[taint.Key]; exists {
+			klog.V(4).Infof("Removing node condition taint %s, when creating template from node %s", taint.Key, node.Name)
+			continue
+		}
+
+		if exists := ignoredTaints[taint.Key]; exists {
+			klog.V(4).Infof("Removing ignored taint %s, when creating template from node %s", taint.Key, node.Name)
+			continue
+		}
+
+		newTaints = append(newTaints, taint)
 	}
 	newNode.Spec.Taints = newTaints
 	return newNode, nil
@@ -525,38 +549,6 @@ func fixNodeGroupSize(context *context.AutoscalingContext, clusterStateRegistry
 	return fixed, nil
 }
 
-// getPotentiallyUnneededNodes returns nodes that are:
-// - managed by the cluster autoscaler
-// - in groups with size > min size
-func getPotentiallyUnneededNodes(context *context.AutoscalingContext, nodes []*apiv1.Node) []*apiv1.Node {
-	result := make([]*apiv1.Node, 0, len(nodes))
-
-	nodeGroupSize := getNodeGroupSizeMap(context.CloudProvider)
-
-	for _, node := range nodes {
-		nodeGroup, err := context.CloudProvider.NodeGroupForNode(node)
-		if err != nil {
-			klog.Warningf("Error while checking node group for %s: %v", node.Name, err)
-			continue
-		}
-		if nodeGroup == nil || reflect.ValueOf(nodeGroup).IsNil() {
-			klog.V(4).Infof("Skipping %s - no node group config", node.Name)
-			continue
-		}
-		size, found := nodeGroupSize[nodeGroup.Id()]
-		if !found {
-			klog.Errorf("Error while checking node group size %s: group size not found", nodeGroup.Id())
-			continue
-		}
-		if size <= nodeGroup.MinSize() {
-			klog.V(1).Infof("Skipping %s - node group min size reached", node.Name)
-			continue
-		}
-		result = append(result, node)
-	}
-	return result
-}
-
 func hasHardInterPodAffinity(affinity *apiv1.Affinity) bool {
 	if affinity == nil {
 		return false
@@ -616,19 +608,6 @@ func getNodeResource(node *apiv1.Node, resource apiv1.ResourceName) int64 {
 	return nodeCapacityValue
 }
 
-func getNodeGroupSizeMap(cloudProvider cloudprovider.CloudProvider) map[string]int {
-	nodeGroupSize := make(map[string]int)
-	for _, nodeGroup := range cloudProvider.NodeGroups() {
-		size, err := nodeGroup.TargetSize()
-		if err != nil {
-			klog.Errorf("Error while checking node group size %s: %v", nodeGroup.Id(), err)
-			continue
-		}
-		nodeGroupSize[nodeGroup.Id()] = size
-	}
-	return nodeGroupSize
-}
-
 // UpdateClusterStateMetrics updates metrics related to cluster state
 func UpdateClusterStateMetrics(csr *clusterstate.ClusterStateRegistry) {
 	if csr == nil || reflect.ValueOf(csr).IsNil() {
